#!/usr/bin/perl

###############################################################################
#
# This file can be distributed and modified freely, as long as this original
# header is not modified. This file is GNU licensed. See COPYING
#
# Author:         Nicolas Hatier <email removed for anti-spam purpose,
#                                  Available on http://www.niversoft.com>
# Organization:   Niversoft ides logicielles / software ideas
# Date:           May 2003
#
# This code is provided AS IS, without any certification that it will even run
#
###############################################################################
#

use Getopt::Long qw/:config ignore_case_always/;
use MIME::Tools;
use MIME::Parser;
use MIME::Decoder::QuotedPrint;
use MIME::Decoder::Base64;
use MIME::Decoder::Binary;
use MIME::Decoder::Gzip64;
use MIME::Decoder::NBit;
use MIME::Decoder::UU;
use MIME::Words qw(:all);
use IPC::Open2;
use strict;

my $VERSION = "1.0.3";

Getopt::Long::Configure("prefix_pattern=-|--|/");
$|=1;

my $Unique = 0;

my %MimeType = (
      ".ai"  => "application/postscript",
     ".aif"  => "audio/x-aiff",
    ".aifc"  => "audio/x-aiff",
    ".aiff"  => "audio/x-aiff",
     ".asc"  => "text/plain",
      ".au"  => "audio/basic",
     ".avi"  => "video/x-msvideo",
   ".bcpio"  => "application/x-bcpio",
     ".bin"  => "application/octet-stream",
       ".c"  => "text/plain",
      ".cc"  => "text/plain",
    ".ccad"  => "application/clariscad",
     ".cdf"  => "application/x-netcdf",
   ".class"  => "application/octet-stream",
    ".cpio"  => "application/x-cpio",
     ".cpt"  => "application/mac-compactpro",
     ".csh"  => "application/x-csh",
     ".css"  => "text/css",
     ".dcr"  => "application/x-director",
     ".dir"  => "application/x-director",
     ".dms"  => "application/octet-stream",
     ".doc"  => "application/msword",
     ".drw"  => "application/drafting",
     ".dvi"  => "application/x-dvi",
     ".dwg"  => "application/acad",
     ".dxf"  => "application/dxf",
     ".dxr"  => "application/x-director",
     ".eps"  => "application/postscript",
     ".etx"  => "text/x-setext",
     ".exe"  => "application/octet-stream",
      ".ez"  => "application/andrew-inset",
       ".f"  => "text/plain",
     ".f90"  => "text/plain",
     ".fli"  => "video/x-fli",
     ".gif"  => "image/gif",
    ".gtar"  => "application/x-gtar",
      ".gz"  => "application/x-gzip",
       ".h"  => "text/plain",
     ".hdf"  => "application/x-hdf",
      ".hh"  => "text/plain",
     ".hqx"  => "application/mac-binhex40",
     ".htm"  => "text/html",
    ".html"  => "text/html",
     ".ice"  => "x-conference/x-cooltalk",
     ".ief"  => "image/ief",
    ".iges"  => "model/iges",
     ".igs"  => "model/iges",
     ".ips"  => "application/x-ipscript",
     ".ipx"  => "application/x-ipix",
     ".jpe"  => "image/jpeg",
    ".jpeg"  => "image/jpeg",
     ".jpg"  => "image/jpeg",
      ".js"  => "application/x-javascript",
     ".kar"  => "audio/midi",
   ".latex"  => "application/x-latex",
     ".lha"  => "application/octet-stream",
     ".lsp"  => "application/x-lisp",
     ".lzh"  => "application/octet-stream",
       ".m"  => "text/plain",
     ".man"  => "application/x-troff-man",
      ".me"  => "application/x-troff-me",
    ".mesh"  => "model/mesh",
     ".mid"  => "audio/midi",
    ".midi"  => "audio/midi",
     ".mif"  => "application/vnd.mif",
    ".mime"  => "www/mime",
     ".mov"  => "video/quicktime",
   ".movie"  => "video/x-sgi-movie",
     ".mp2"  => "audio/mpeg",
     ".mp3"  => "audio/mpeg",
     ".mpe"  => "video/mpeg",
    ".mpeg"  => "video/mpeg",
     ".mpg"  => "video/mpeg",
    ".mpga"  => "audio/mpeg",
      ".ms"  => "application/x-troff-ms",
     ".msh"  => "model/mesh",
      ".nc"  => "application/x-netcdf",
     ".oda"  => "application/oda",
     ".pbm"  => "image/x-portable-bitmap",
     ".pdb"  => "chemical/x-pdb",
     ".pdf"  => "application/pdf",
     ".pgm"  => "image/x-portable-graymap",
     ".pgn"  => "application/x-chess-pgn",
     ".png"  => "image/png",
     ".pnm"  => "image/x-portable-anymap",
     ".pot"  => "application/mspowerpoint",
     ".ppm"  => "image/x-portable-pixmap",
     ".pps"  => "application/mspowerpoint",
     ".ppt"  => "application/mspowerpoint",
     ".ppz"  => "application/mspowerpoint",
     ".pre"  => "application/x-freelance",
     ".prt"  => "application/pro_eng",
      ".ps"  => "application/postscript",
      ".qt"  => "video/quicktime",
      ".ra"  => "audio/x-realaudio",
     ".ram"  => "audio/x-pn-realaudio",
     ".ras"  => "image/cmu-raster",
     ".rgb"  => "image/x-rgb",
      ".rm"  => "audio/x-pn-realaudio",
    ".roff"  => "application/x-troff",
     ".rpm"  => "audio/x-pn-realaudio-plugin",
     ".rtf"  => "text/rtf",
     ".rtx"  => "text/richtext",
     ".scm"  => "application/x-lotusscreencam",
     ".set"  => "application/set",
     ".sgm"  => "text/sgml",
    ".sgml"  => "text/sgml",
      ".sh"  => "application/x-sh",
    ".shar"  => "application/x-shar",
    ".silo"  => "model/mesh",
     ".sit"  => "application/x-stuffit",
     ".skd"  => "application/x-koan",
     ".skm"  => "application/x-koan",
     ".skp"  => "application/x-koan",
     ".skt"  => "application/x-koan",
     ".smi"  => "application/smil",
    ".smil"  => "application/smil",
     ".snd"  => "audio/basic",
     ".sol"  => "application/solids",
     ".spl"  => "application/x-futuresplash",
     ".src"  => "application/x-wais-source",
    ".step"  => "application/STEP",
     ".stl"  => "application/SLA",
     ".stp"  => "application/STEP",
 ".sv4cpio"  => "application/x-sv4cpio",
  ".sv4crc"  => "application/x-sv4crc",
     ".swf"  => "application/x-shockwave-flash",
       ".t"  => "application/x-troff",
     ".tar"  => "application/x-tar",
     ".tcl"  => "application/x-tcl",
     ".tex"  => "application/x-tex",
    ".texi"  => "application/x-texinfo",
 ".texinfo"  => "application/x-texinfo",
     ".tif"  => "image/tiff",
    ".tiff"  => "image/tiff",
      ".tr"  => "application/x-troff",
     ".tsi"  => "audio/TSP-audio",
     ".tsp"  => "application/dsptype",
     ".tsv"  => "text/tab-separated-values",
     ".txt"  => "text/plain",
     ".unv"  => "application/i-deas",
   ".ustar"  => "application/x-ustar",
     ".vcd"  => "application/x-cdlink",
     ".vda"  => "application/vda",
     ".viv"  => "video/vnd.vivo",
    ".vivo"  => "video/vnd.vivo",
    ".vrml"  => "model/vrml",
     ".wav"  => "audio/x-wav",
     ".wrl"  => "model/vrml",
     ".xbm"  => "image/x-xbitmap",
     ".xlc"  => "application/vnd.ms-excel",
     ".xll"  => "application/vnd.ms-excel",
     ".xlm"  => "application/vnd.ms-excel",
     ".xls"  => "application/vnd.ms-excel",
     ".xlw"  => "application/vnd.ms-excel",
     ".xml"  => "text/xml",
     ".xpm"  => "image/x-xpixmap",
     ".xwd"  => "image/x-xwindowdump",
     ".xyz"  => "chemical/x-pdb",
     ".zip"  => "application/zip"
);

my $rundir = $0;
$rundir =~ s|\\|/|g;
$rundir = index($rundir, "/")>=0 ? substr($rundir, 0, rindex($rundir, "/")) : ".";
my $TNEF = "$rundir/tnef";
my $TEMP = "$rundir/tneftmp";

main();

sub ReplaceAttachements
{
	my @newparts = ();
	my $changed = 0;

	foreach my $part ($_[0]->parts)
	{
		if ($part->head->recommended_filename() =~ /winmail.dat/i)
		{
         my $winmail_dat_file = $part->bodyhandle->path();

			my $pid = open2(\*RDRFH, \*WTRFH, $TNEF, '-f', $winmail_dat_file, "-C", $TEMP, '--number-backups');

			foreach my $file (<RDRFH>)
			{

            $file =~ s/[\r\n]//g;
				return -1 if ($file =~ /\*\*\*/);
            my @ext = ($file =~ /(\.\w+)$/);
				my $ctype = $MimeType{$ext[0]};
				$ctype = "application/octet-stream" if (!defined $ctype);
            my @fname = $file =~ m|$TEMP/(.*)|;
				print "* Replacing @fname[0] ($ctype)\n";
   			push @newparts, MIME::Entity->build(Path => "$file", Type => $ctype, Encoding => "-SUGGEST");
   		}
   		wait();
			$changed = 1;
		}
		else
		{
			push @newparts, $part;
		}
	}
	$_[0]->parts(\@newparts) if ($changed);
	return $changed;
}

sub CleanTemp()
{
	opendir DIR, $TEMP;
	unlink "$TEMP/$_" foreach (readdir(DIR));
	closedir DIR;
}

sub ProcessDir
{
	print <<END and return if (!-f $TNEF);
Unable to find the tnef binary. Please follow the instruction in INSTALL.
tnef must be in the same directory than WinMailDatFilter.pl
END

	my $DirectoryToProcess = shift;
	my $Recurse = not shift;

	opendir DIR, $DirectoryToProcess;
	foreach my $CurrentEntry (grep {/^[^\.]/} readdir(DIR))
	{
		my $EntityPath = "$DirectoryToProcess/$CurrentEntry";

		if (-d $EntityPath)
		{
			next unless($Recurse);
			print "---- $CurrentEntry\n";
			ProcessDir($EntityPath);
		}
		else
		{
			next if ($CurrentEntry =~ /\./);
			CleanTemp();
			open FILE, "$EntityPath";
			my $parser = new MIME::Parser;
			$parser->output_dir($TEMP);
			$parser->filer->ignore_filename(1);
			my $entity = $parser->parse(\*FILE);

			close FILE;
			if (ReplaceAttachements($entity) == 1)
			{
				open FILE, ">$EntityPath";
				binmode FILE if ($^O !~ /MSWin32/i);
			 	$entity->print(\*FILE);
			 	close FILE;
			}
			$parser->filer->purge();
		}
	}
	closedir DIR;
}

sub main
{
   my %params;
   my @options = ("path|p=s", "help|?|h", "norecurse|n|N");
   my $ValidCommand = GetOptions(\%params, @options);
   $TNEF .= ".exe" if ($^O =~ /Win32/);

   mkdir $TEMP;
   if (!$ValidCommand || $params{help} || ($params{norecurse} and !defined $params{path}))
   {
      print <<END;

Niversoft's WinMailDatFilter, external filter/attachment replacer.
Version $VERSION

   Replaces winmail.dat attachments in mail (RFC822) messages with the real
   attachments encapsuled in this file. Can work as a Communigate Pro External
   Filter, or can be used to convert existing mailbox.

   Usage: WinMailDatFilter [-p|--path=mailboxpath] [-n|--norecurse] [-h|-help]

   When called with no parameter, starts in External Filter mode

   -p, --path      :  specifies the path to the account/mailbox to convert.
                      The mailbox must be in .mdir format (MailDirMailBox).
                      Only files with no extension will be converted.
                      Files are converted in place, so you are strongly
                      encouraged to make a backup before running.
                      Directories are processed recursively, unless the
                      norecurse flag is specified

   -n, --norecurse :  Don't precess directories recursively, only process the
                      specified mailbox.

   -h, -?, --help  :  Shows this help message

   There is some cases when the winmail.dat file can't be parsed (file may be
   corrupted, etc.) In those case, the message is left unchanged.

   Processing time varies with the CPU and the disk speed. Average time is
   two messages per second.

   In EXTFILTER mode, you are encouraged to set a rule in Communigate Pro to
   make WinMailDatFilter filter only messages generated by Microsoft Outlook,
   This way, your server won't be slowed down for unneeded filtering. The
   rule condition is the following:
      Header Field  is  X-Mailer: Microsoft Outlook*

   Contact info\@niversoft.com for any question.

   GNU Licensed
      http://www.gnu.org/licenses/gpl.txt
   Uses slightly modified TNEF
      http://sourceforge.net/projects/tnef
   Uses MIME::Tools
      http://search.cpan.org/dist/MIME-tools
   All modified sources available on
      http://www.niversoft.com/products/cgscripts/cgscripts.html

END
   }
   elsif (defined $params{path})
   {
		ProcessDir($params{path}, $params{norecurse});
   }
   else
   {
      ExtFilter();
   }

   CleanTemp();
   rmdir $TEMP;

}

sub ExtFilter
{
	print <<END and return if (!-f $TNEF);
* Unable to find the tnef binary. Please follow the instruction in INSTALL.
* tnef must be in the same directory than WinMailDatFilter.pl
END


   print <<END;
*
* Niversoft's WinMailDatFilter, external filter/attachment replacer running...
* Version $VERSION
*
* GNU Licensed, see http://www.gnu.org/licenses/gpl.txt
* Uses slightly modified TNEF, see http://sourceforge.net/projects/tnef
* Uses MIME::Tools, see http://search.cpan.org/dist/MIME-tools
* All modified sources available on http://www.niversoft.com/products/cgscripts/cgscripts.html
*
* Temp Directory: $TEMP
* TNEF path: $TNEF
*
END

   while (<>)
   {
      my ($ID, $Command, $File) = split / |\r|\n/;

      print "$ID OK\n"      and next if ($Command =~ /^quit$/i);
      print "$ID OK\n"      and next if ($Command =~ /^key$/i);
	   print "$ID INTF 3\n"  and next if ($Command =~ /^intf$/i);
		print "$ID FAILURE\n" and next if ($Command !~ /^file$/i); # other unsupported commands

      CleanTemp();
      my $Sender;
      my @Recipients;
      if (open(FILE, $File))
      {
         my @exts;
         {
            my $line = <FILE>;
            if ($line =~ /^\w \w+[ \r\n]/)
            {
               chomp($line);
               while($line ne '')
               {
                  if($line =~ /^(\w).+<(.+)>/)
                  {
                     $Sender = $2 if($1 eq 'P');
                     push @Recipients, $2 if($1 eq 'R')
                  }
                  $line = <FILE>;
                  chomp($line);
               }
            }
            else
            {
               seek FILE, 0, 0;
            }

			   my $parser = new MIME::Parser;
			   $parser->output_dir($TEMP);
			   $parser->filer->ignore_filename(1);
			   my $entity = $parser->parse(\*FILE);
			   close FILE;

            my $resp = ReplaceAttachements($entity);
			   if ($resp == 1)
			   {
               my $file = "Submitted/wmdtmp" . time() . ($Unique++);
               if (!open FILE, ">$file.tmp")
               {
                  print "* Unable to write output message. Message is left unchanged\n";
                  print "$ID ADDHEADER \"X-Filter-Report: Unable to write output message.\"\n";
               }
               else
               {
                  print "$ID DISCARD\n";
                  print FILE "Return-Path: $Sender\n";
                  print FILE "Envelope-To: $_\n" foreach (@Recipients);
                  print FILE "X-WinMailDatFilter: Attachments Processed\n";
			 	      $entity->print(\*FILE);
			 	      close FILE;
                  rename "$file.tmp", "$file.sub";
               }
			   }
            elsif ($resp == 0)
            {
               print "$ID OK\n";
            }
            elsif ($resp == 0)
            {
               print "* Unable to decode winmail.dat in this message. Message is left unchanged\n";
               print "$ID ADDHEADER \"X-Filter-Report: Unable to decode winmail.dat\"\n";
            }
			   $parser->filer->purge();
         }

      }
      else
      {
            print "* Error file not found\n";
            print "$ID OK\n";
      }
      open STDOUT, ">&STDOUT";
   }
}
